from matplotlib.colors import Normalize, LogNorm
from dataclasses import dataclass
import numpy as np

from .RowCollectorClass import RowCollector

@dataclass
class Ranges:
    """ Dataclass that contains data ranges
    """
    minpos: float
    min: float
    max: float

class NormalizeData:
    """ Normalize numerical data across multiple plots
    """

    xaxis: bool
    yaxis: bool
    _collector: RowCollector
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, tb):
        pass

    def __init__(self, xaxis=None, yaxis=None):
        self.xaxis = xaxis
        self.yaxis = yaxis
        columns = ['zminpos','zmin','zmax']
        if xaxis: columns += ['xminpos','xmin','xmax']
        if yaxis: columns += ['yminpos','ymin','ymax']
        self._collector = RowCollector(columns)

    def append(self, vdata, xdata=None, ydata=None):
        """ Append data to the range collector
        
        Statistics of x and y axes is optional

        :param vdata: Numerical data
        :param xdata: X-axis points
        :param ydata: Y-axis points
        """
        zminpos = np.nanmin(vdata[vdata>0]) if np.sum(vdata>0) else np.nan
        ranges = [zminpos,np.min(vdata),np.max(vdata)]
        if self.xaxis:
            if xdata is None:
                raise Exception("Missing x-axes values")
            xminpos = np.nanmin(xdata[xdata>0]) if np.sum(xdata>0) else np.nan
            ranges += [xminpos,np.min(xdata),np.max(xdata)]
        if self.yaxis:
            if ydata is None:
                raise Exception("Missing y-axes values")
            yminpos = np.nanmin(ydata[ydata>0]) if np.sum(ydata>0) else np.nan
            ranges += [yminpos,np.min(ydata),np.max(ydata)]
        self._collector.append(ranges)

    def data(self):
        """ Return collected ranges as a dictionary
        """
        return self._collector.to_dict()

    def xranges(self):
        """ Return x-axis ranges
        """
        return Ranges(
            minpos = np.nanmin(self._collector.xminpos),
            min = np.nanmin(self._collector.xmin),
            max = np.nanmax(self._collector.xmax)
        )
    
    def yranges(self):
        """ Return y-axis ranges
        """
        return Ranges(
            minpos = np.nanmin(self._collector.yminpos),
            min = np.nanmin(self._collector.ymin),
            max = np.nanmax(self._collector.ymax)
        )
        
    def zranges(self):
        """ Return data ranges
        """
        return Ranges(
            minpos = np.nanmin(self._collector.zminpos),
            min = np.nanmin(self._collector.zmin),
            max = np.nanmax(self._collector.zmax)
        )
    
    def linnorm(self):
        """ Return linear norm from ranges
        """
        return Normalize(
            vmin=np.nanmin(self._collector.zmin), 
            vmax=np.nanmax(self._collector.zmax)
        )

    def lognorm(self):
        """ Return logarithmic norm from ranges
        """
        return LogNorm(
            vmin=np.log10(np.nanmin(self._collector.zminpos)), 
            vmax=np.log10(np.nanmax(self._collector.zmax))
        )



import numpy as np
import pandas as pd
from textwrap import dedent
import pytest
import sys
sys.path.insert(0, 'src')

import scinumtools as snt

@pytest.fixture
def rows():
    return [[1,2,3],
            [4,5,6],
            [7,8,0]]
            
@pytest.fixture
def columns():
    return ['col1','col2','col3']

def test_append_list(columns, rows):

    def check_values(rc):
        assert rc.size() == len(rows)
        assert rc.col1 == [1,4,7]
        assert rc.col2 == [2,5,8]
        assert rc.col3 == [3,6,0]

    with snt.RowCollector(columns) as rc:
        for row in rows:
            rc.append(row)
        check_values(rc)
    with snt.RowCollector(columns,rows) as rc:
        check_values(rc)

def test_conversions_list(columns,rows):
    
    with snt.RowCollector(columns,rows) as rc:
        assert rc.to_dict() == dict(
            col1 = [1,4,7],
            col2 = [2,5,8],
            col3 = [3,6,0],
        )
        pd.testing.assert_frame_equal(
            rc.to_dataframe(),
            pd.DataFrame(dict(
                col1 = [1,4,7],
                col2 = [2,5,8],
                col3 = [3,6,0],
            ))
        )
        result = dedent("""\
               col1  col2  col3
            0     1     2     3
            1     4     5     6
            2     7     8     0
        """.rstrip())
        assert rc.to_text() == result
        assert str(rc) == result

def test_append_array(columns, rows):
    
    def check_values(rc):
        assert rc.size() == len(rows)
        np.testing.assert_equal(rc.col1, [1,4,7])
        np.testing.assert_equal(rc.col2, [2,5,8])
        np.testing.assert_equal(rc.col3, [3,6,0])
        
    with snt.RowCollector(columns, array=True) as rc:
        for row in rows:
            rc.append(row)
        check_values(rc)
    with snt.RowCollector(columns, rows, array=True) as rc:
        check_values(rc)

def test_conversions_array(columns, rows):
    
    with snt.RowCollector(columns, rows, array=True) as rc:
        np.testing.assert_equal(rc.to_dict(), dict(
            col1 = [1,4,7], 
            col2 = [2,5,8],
            col3 = [3,6,0],
        ))
        pd.testing.assert_frame_equal(
            rc.to_dataframe(),
            pd.DataFrame(dict(
                col1 = [1.,4.,7.],
                col2 = [2.,5.,8.],
                col3 = [3.,6.,0.],
            ))
        )
        result = dedent("""\
               col1  col2  col3
            0   1.0   2.0   3.0
            1   4.0   5.0   6.0
            2   7.0   8.0   0.0
        """.rstrip())
        assert rc.to_text() == result
        assert str(rc) == result
        
    columns = {'col1':dict(dtype=str),'col2':dict(dtype=float),'col3':dict(dtype=bool)}
    with snt.RowCollector(columns, rows, array=True) as rc:
        np.testing.assert_equal(rc.col1, ['1','4','7'])
        np.testing.assert_equal(rc.col2, [2,5,8])
        np.testing.assert_equal(rc.col3, [True,True,False])
        np.testing.assert_equal(rc.to_dict(), dict(
            col1 = ['1','4','7'], 
            col2 = [2,5,8],
            col3 = [True,True,False],
        ))
        pd.testing.assert_frame_equal(
            rc.to_dataframe(),
            pd.DataFrame(dict(
                col1 = ['1','4','7'],
                col2 = [2.,5.,8.],
                col3 = [True,True,False],
            ))
        )
        result = dedent("""\
              col1  col2   col3
            0    1   2.0   True
            1    4   5.0   True
            2    7   8.0  False
        """.rstrip())
        assert rc.to_text() == result
        assert str(rc) == result



from typing import Union

from .OperatorClass import *
from .ExpressionClass import Expression
from .TokensClass import Tokens

class ExpressionSolver:

    tokens: list
    operators: list
    expr: Expression
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, tb):
        pass
    
    def __init__(self, atom, operators:list = None):
        self.tokens = Tokens(atom)
        self.operators = operators if operators else [
            OperatorLog, OperatorLog10, OperatorLogb,
            OperatorExp, OperatorSqrt, OperatorPowb,
            OperatorSin, OperatorCos, OperatorTan,
            OperatorPar,  # should be the last of parenthesis operators
            OperatorPow,
            OperatorMul, OperatorTruediv,
            OperatorAdd, OperatorSub,
            OperatorEq, OperatorNe,
            OperatorNot,  # needs to be after OperatorNe
            OperatorLe, OperatorGe,
            OperatorLt, OperatorGt,
            OperatorAnd, OperatorOr, 
        ]

    def solve(self, expr:Union[str,Expression]):
        self.expr = Expression(expr) if isinstance(expr, str) else expr
        
        # Tokenize expression
        while self.expr.right:
            for operator in self.operators:
                if self.expr.right.startswith(operator.symbol):
                    # Create atom from the left side and append it to tokens
                    if left:=self.expr.pop_left():
                        self.tokens.append(self.tokens.atom(left))
                    # Initialize an operator
                    op = operator(self.expr)
                    if op.args:
                        # Solve operator arguments
                        with ExpressionSolver(self.tokens.atom) as es:
                            for a in range(len(op.args)):
                                op.args[a] = es.solve(op.args[a])
                    # Append operator to tokens
                    self.tokens.append(op)
                    break
            else:
                self.expr.shift()
        # Create atom from the remaining left side
        if left:=self.expr.pop_left():
            self.tokens.append(self.tokens.atom(left))

            
        # Perform operation steps
        osteps = [
            dict(operators=[
                OperatorLog, OperatorLog10, OperatorLogb,
                OperatorExp, OperatorSqrt, OperatorPowb,
                OperatorSin, OperatorCos, OperatorTan,
                OperatorPar
            ], otype=Otype.ARGS),
            dict(operators=[OperatorAdd, OperatorSub],     otype=Otype.UNARY),
            dict(operators=[OperatorPow],                  otype=Otype.BINARY),
            dict(operators=[OperatorMul, OperatorTruediv], otype=Otype.BINARY),
            dict(operators=[OperatorAdd, OperatorSub],     otype=Otype.BINARY),
            dict(operators=[
                OperatorEq, OperatorNe,
                OperatorLe, OperatorGe,
                OperatorLt, OperatorGt,
            ], otype=Otype.BINARY),
            dict(operators=[OperatorNot],                  otype=Otype.UNARY),
            dict(operators=[OperatorAnd],                  otype=Otype.BINARY),
            dict(operators=[OperatorOr],                   otype=Otype.BINARY),
        ]
        for o,ostep in enumerate(osteps):
            ostep['operators'] = tuple([o for o in ostep['operators'] if o in self.operators])
            if ostep['operators']:
                self.tokens.operate(**ostep)

        # Return the final atom
        if len(self.tokens.left)>0 or len(self.tokens.right)>1:
            raise Exception("Cannot solve expression due to unprocessed tokens:", self.tokens.left, self.tokens.right)
        atom = self.tokens.get_right()
        #print(self.expr.expr, atom)
        return atom



from .AtomClass import AtomBase
from .OperatorClass import Otype

class Tokens:
    
    atom: AtomBase
    left: list
    right: list
    
    def __init__(self, atom):
        self.atom = atom
        self.left = []
        self.right = []
        
    def append(self, token):
        self.right.append(token)
        
    def get_left(self):
        return self.left.pop() if self.left else None
    
    def get_right(self):
        return self.right.pop(0) if self.right else None
    
    def put_left(self, token):
        self.left.append(token)
        
    def put_right(self, token):
        self.right.insert(0, token)
        
    def operate(self, operators, otype):
        iterator = 0
        while self.right:
            token = self.right.pop(0)
            if isinstance(token, operators) and otype==Otype.UNARY:
                token.operate_unary(self)
            elif isinstance(token, operators) and otype==Otype.BINARY:
                token.operate_binary(self)
            elif isinstance(token, operators) and otype==Otype.ARGS:
                token.operate_args(self)
            else:
                self.put_left(token)
        self.right = self.left
        self.left = []



import numpy as np
from dataclasses import dataclass, field, fields
from typing import Union

from .settings import *

@dataclass
class Fraction:

    num: Union[int, tuple, str]   # numerator
    den: int = field(default=1)   # denominator

    def __post_init__(self):
        # initialize from a tuple
        if isinstance(self.num, str):
            if SYMBOL_FRACTION in self.num:
                num, den = self.num.split(SYMBOL_FRACTION)
                self.num, self.den = int(num), int(den)
            else:
                self.num, self.den = int(self.num), 1
        elif isinstance(self.num, tuple):
            self.num, self.den = self.num
        elif isinstance(self.num, Fraction):
            value = self.num
            self.num = value.num
            self.den = value.den
        # ensure correct type
        if isinstance(self.num, (int,float)) and isinstance(self.den, (int,float)):
            self.num = int(self.num)
            self.den = int(self.den)
        else:
            raise Exception("Incorrect numerator, or denominator:", self.num, self.den)
        # enforce whole numbers
        if np.mod(self.num,1)!=0 or np.mod(self.den,1)!=0:
            raise Exception("Fraction excepts only whole numbers:", self.num, self.den)
        else:
            self.num = int(self.num)
            self.den = int(self.den)
        # reset zero values
        if self.num in [0, -0]:
            self.num = 0
            self.den = 1
        # keep minus sign always on the top
        if self.num>=0 and self.den<0:
            self.num = -self.num
            self.den = -self.den
        elif self.num<0 and self.den<0:
            self.num = -self.num
            self.den = -self.den
        # remove common divisors
        def reduce(num: int, den:int):
            gcd=np.gcd(num, den)
            if gcd>1:
                return reduce(int(num/gcd), int(den/gcd))
            return int(num), int(den)
        self.num, self.den = reduce(self.num, self.den)
    
    def __str__(self):
        if self.num==0 or self.den==1:
            return str(self.num)
        else:
            return f"{self.num}{SYMBOL_FRACTION}{self.den}"

    def __repr__(self):
        if self.num==0 or self.den==1:
            return str(self.num)
        else:
            return f"Fraction({self.num},{self.den})"        
        
    def __add__(self, other):
        if not isinstance(other, Fraction):
            other = Fraction(other)
        return Fraction(self.num*other.den+other.num*self.den, self.den*other.den)
            
    def __sub__(self, other):
        if not isinstance(other, Fraction):
            other = Fraction(other)
        return Fraction(self.num*other.den-other.num*self.den, self.den*other.den)

    def __mul__(self, other):
        if isinstance(other, Fraction):
            return Fraction(self.num*other.num, self.den*other.den)
        elif isinstance(other, tuple):
            return Fraction(self.num*other[0], self.den*other[1])
        else:
            return Fraction(self.num*other, self.den)

    def __truediv__(self, other):
        if isinstance(other, Fraction):
            return Fraction(self.num*other.den, self.den*other.num)
        elif isinstance(other, tuple):
            return Fraction(self.num*other[1], self.den*other[0])
        else:
            return Fraction(self.num, self.den*other)
    
    def __neg__(self):
        return Fraction(-self.num, self.den)
    
    def value(self, dtype=tuple):
        if self.num==0 or self.den==1:
            return self.num
        elif dtype==tuple:
            return (self.num,self.den)
        elif dtype==float:
            return self.num/self.den



from dataclasses import dataclass
import numpy as np
import pandas as pd
from typing import Union

@dataclass
class RowCollector:
    """ RowCollector class collect table rows and transform them into one of available formats

    Example of use:

    ..code-block::
    
        import scinumtools as snt
        
        with snt.structs.RowCollector(['col1','col2','col3']) as rc:
             rc.append([1,2,3])
             rc.append([4,5,6])
             data = rc.to_dict()
    
    :param columns: This can be either a list of column names or a dictionary with column array settings
    :param bool array: Stores values as numpy arrays if 'True', otherwise store data in lists
    """
    _columns: list
    _array: bool
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, tb):
        pass
    
    def __init__(self, columns: Union[list,dict], rows: list = None, array: bool = False):
        self._columns = []
        self._array = array
        if self._array is False:
            self._columns = []
            for column in columns:      
                setattr(self,column,[])
                self._columns.append(column)
        else:
            if isinstance(columns,dict):
                for column, kwargs in columns.items():
                    setattr(self,column,np.array([],**kwargs))
                    self._columns.append(column)
            else:
                for column in columns:
                    setattr(self,column,np.array([]))
                    self._columns.append(column)
        if rows:
            for row in rows:
                self.append(row)

    def __str__(self):
        return self.to_text()
    
    def append(self, values: list):
        """ Append a single row

        :param values: List of values for each column
        """
        if self._array:
            for n, name in enumerate(self._columns):
                data = getattr(self,name)
                new = np.array(values[n],dtype=data.dtype)
                setattr(self,name, np.append(data,new) )
        else:
            for n, name in enumerate(self._columns):
                getattr(self,name).append(values[n])
            
    def sort(self, name: str, reverse=False):
        """ Sort all columns according to one

        :param str name: name of the sorting column
        """
        ids = np.argsort(getattr(self, name))
        if reverse: ids = ids[::-1]
        if self._array:
            for n, name in enumerate(self._columns):
                setattr(self,name,getattr(self, name)[ids])
        else:
            for n, name in enumerate(self._columns):
                setattr(self,name,list(np.array(getattr(self, name))[ids]))
            
    def size(self):
        """ Get number of items in columns
        """
        if self._columns:
            return len(getattr(self, self._columns[0]))
        else:
            return 0
        
    def to_dict(self):
        """ Convert class data to a dictionary of lists/arrays
        """
        data = {}
        for name in self._columns:
            data[name] = getattr(self,name)
        return data
    
    def to_dataframe(self, columns: Union[list,dict]=None):
        """ Convert class data to a pandas data frame

        :param columns: This can be either a list of columns or a dictionary of column:title pairs. If not set, all coumns are being taken.
        """
        if isinstance(columns,dict):
            return pd.DataFrame({title:getattr(self,name) for name,title in columns.items()})
        elif isinstance(columns,list):
            return pd.DataFrame({name:getattr(self,name) for name in columns})
        else:
            return pd.DataFrame({name:getattr(self,name) for name in self._columns})
        
    def to_text(self, **kwargs):
        """ Convert class data to a text using pandas dataframe

        :param kwargs: kwargs of DataFrame's to_string() method
        """
        return self.to_dataframe().to_string(**kwargs)



from .QuantityClass import Quantity
from .QuantityClass  import Quantity as quant
from .UnitClass import Unit
from .UnitClass import Unit as unit
from .ConstantClass import Constant
from .ConstantClass import Constant as const
from .NaNClass import NaN
from .NaNClass import NaN as nan
from .MagnitudeClass import Magnitude
from .DimensionsClass import Dimensions
from .FractionClass import Fraction
from .BaseUnitsClass import BaseUnits
from .UnitSolver import UnitSolver


import numpy as np
from typing import Union

class DataPlotGrid:
    data: Union[list,dict]
    ndata: int
    ncols: int
    nrows: int
    figsize: tuple

    def __init__(self, data: Union[list,dict], ncols:int=2, axsize:tuple=(4,2)):
        self.data = data
        self.ndata = len(data)
        self.ncols = ncols
        self.nrows = int(np.ceil(self.ndata/self.ncols))
        self.figsize = (self.ncols*axsize[0], self.nrows*axsize[1])

    def items(self, missing:bool=None, transpose:bool=False):
        if missing:
            for i in range(self.ndata, self.ncols*self.nrows):
                if transpose:
                    yield (i,int(i%self.nrows),int(i/self.nrows))
                else:
                    yield (i,int(i/self.ncols),int(i%self.ncols))
        else:
            if isinstance(self.data, list):
                for i,d in enumerate(self.data):
                    if transpose:
                        yield (i,int(i%self.nrows),int(i/self.nrows),d)
                    else:
                        yield (i,int(i/self.ncols),int(i%self.ncols),d)
            elif isinstance(self.data, dict):
                for i,(k,v) in enumerate(self.data.items()):
                    if transpose:
                        yield (i,int(i%self.nrows),int(i/self.nrows),k,v)
                    else:
                        yield (i,int(i/self.ncols),int(i%self.ncols),k,v)
            else:
                raise Exception('Wrong data type:', self.data)



import numpy as np
import pytest
from math import isclose
import os
import sys
sys.path.insert(0, 'src')

from scinumtools.units import *
    
def test_initialization():

    # Test initialization
    assert str(Fraction(2))   == "2"
    assert str(Fraction(0,3)) == "0"
    assert str(Fraction(1,3)) == "1:3"

def test_arithmetics():
    f1 = Fraction(1,3)
    f2 = Fraction(3,2)
    # with scalars
    assert str(f1+3)        == "10:3"
    assert str(f1-3)        == "-8:3"
    assert str(f1*3)        == "1"
    assert str(f1/3)        == "1:9"
    # with tuples
    assert str(f1+(3,2))    == "11:6"
    assert str(f1-(3,2))    == "-7:6"
    assert str(f1*(3,2))    == "1:2"
    assert str(f1/(3,2))    == "2:9"
    # with fractions
    assert str(f1+f2)       == "11:6"
    assert str(f1-f2)       == "-7:6"
    assert str(f1*f2)       == "1:2"
    assert str(f1/f2)       == "2:9"


import numpy as np
import os
import pytest
import sys
sys.path.insert(0, 'src')

from scinumtools.solver import *

def test_solver_basic():

    expressions = {
        '1+2':           1+2,
        '1-2':           1-2,
        '1*2':           1*2,
        '1/2':           1/2,
        '((2+3) /(3) )': ((2+3) / 3 ),
        "1 * ((2+3) / +3 - -10 ) + (-23 *++2) + 23**2":
        1 * ((2+3) / +3 - -10 ) + (-23 *++2) + 23**2
    }
    with ExpressionSolver(AtomBase) as es:
        for expr, value in expressions.items():
            result = es.solve(expr).value
            assert np.isclose(result, value)
        
def test_solver_math():

    expressions = {
        'exp(23+3)':          np.exp(23+3),
        'log(23+3)':          np.log(23+3),
        'log10(23+3)':        np.log10(23+3),
        'sqrt(23+3)':         np.sqrt(23+3),
        'sin(23+3)':          np.sin(23+3),
        'cos(23+3)':          np.cos(23+3),
        'tan(23+3)':          np.tan(23+3),
        'logb(23+3, 10)':     np.log(23+3)/np.log(10),
        'pow(23+3, 2)':       np.power(23+3, 2),
        '3*1+6/4*sin(34)':    3*1+6/4*np.sin(34),
        'sin(cos(23))':       np.sin(np.cos(23)),
        'pow((23+3), (2*4))': np.power((23+3), (2*4)),
    }
    with ExpressionSolver(AtomBase) as es:
        for expr, value in expressions.items():
            result = es.solve(expr).value
            assert np.isclose(result, value)
            
def test_solver_logical():

    expressions = {
        '1 && 1':    1 and 1,
        '1 && 0':    1 and 0,
        '0 && 0':    0 and 0,
        '1 || 1':    1 or 1,
        '1 || 0':    1 or 0,
        '0 || 0':    0 or 0,
        '!1':        not 1,
        '!0':        not 0,
        '1 && 0 || 1 && !0 && 1 || 0': 1 and 0 or 1 and not 0 and 1 or 0,
        '1-1 && (1*32)/2 || sin(3)':   1-1 and (1*32)/2 or np.sin(3),
    }
    with ExpressionSolver(AtomBase) as es:
        for expr, value in expressions.items():
            result = es.solve(expr)
            assert np.isclose(result.value, value)
            
def test_solver_logical():

    expressions = {
        '1 == 1':    1 == 1,
        '1 == 0':    1 == 0,
        '1 != 1':    1 != 1,
        '1 != 0':    1 != 0,
        '23 <= 45':  23 <= 45,
        '45 <= 45':  45 <= 45,
        '81 <= 45':  81 <= 45,
        '23 >= 45':  23 >= 45,
        '45 >= 45':  45 >= 45,
        '81 >= 45':  81 >= 45,
        '23 < 45':   23 < 45,
        '45 < 45':   45 < 45,
        '81 < 45':   81 < 45,
        '23 > 45':   23 > 45,
        '45 > 45':   45 > 45,
        '81 > 45':   81 > 45,
        'sin(23) < 1 && 3*2 == 6 || !(23 > 43) && cos(0) == 1':
        np.sin(23) < 1 and 3*2 == 6 or not (23 > 43) and np.cos(0) == 1
    }
    with ExpressionSolver(AtomBase) as es:
        for expr, value in expressions.items():
            result = es.solve(expr)
            assert result.value == value

def test_solver_atom():

    foo = 3
    bar = 4
    
    class Atom(AtomBase):
        
        def __init__(self, value:str):
            if value=='foo':
                self.value = foo
            elif value=='bar':
                self.value = bar
            else:
                self.value = float(value)
                
    expressions = {
        'foo < bar && foo * bar == 12':    foo < bar and foo*bar==12,
        'foo*bar**(2-233)':                foo*bar**(2-233),
        'sin(foo)':                        np.sin(foo),
    }
    with ExpressionSolver(Atom) as es:
        for expr, value in expressions.items():
            result = es.solve(expr)
            assert result.value == value
    
def test_solver_operators():

    expressions = {
        '23 > 4':    23 > 4,
    }
    
    operators = {OperatorGt}
    with ExpressionSolver(AtomBase, operators) as es:
        for expr, value in expressions.items():
            result = es.solve(expr)
            assert result.value == value

    operators = {OperatorEq}
    with ExpressionSolver(AtomBase, operators) as es:
        for expr, value in expressions.items():
            with pytest.raises(ValueError) as excinfo:
                result = es.solve(expr)
            assert str(excinfo.value)=="could not convert string to float: '23 > 4'"



import numpy as np
import pytest
from math import isclose
import os
import sys
sys.path.insert(0, 'src')

from scinumtools.units import *

def test_initialization():
    
    assert str(BaseUnits())                 == "BaseUnits()"
    dim = [0,2,3,-2,(3,2),0,0,0]
    assert str(BaseUnits(dim))              == "BaseUnits(g=2 s=3 K=-2 C=3:2)"
    assert str(BaseUnits(Dimensions(*dim))) == "BaseUnits(g=2 s=3 K=-2 C=3:2)"
    assert str(BaseUnits("kg*m2/s2"))       == "BaseUnits(kg=1 m=2 s=-2)"
    assert str(BaseUnits(BaseUnits("kg")))  == "BaseUnits(kg=1)"

def test_arithmetics():
    
    bu1 = BaseUnits({'k:m':3,'g':(3,2)})
    bu2 = BaseUnits({'k:m':2,'g':(4,7)})
    assert not bu1 == bu2
    assert str(bu1+bu2)  == "BaseUnits(km=5 g=29:14)"
    assert str(bu1-bu2)  == "BaseUnits(km=1 g=13:14)"
    assert str(bu1*2)    == "BaseUnits(km=6 g=3)"
    assert str(bu2*0.5)  == "BaseUnits(km=1 g=2:7)"

def test_values():
    # Test scalars and factors
    value = {"k:m": 2, "K": (3,2)}
    bu = BaseUnits(dict(value))
    assert str(bu) == "BaseUnits(km=2 K=3:2)"
    assert bu.value() == value
    
    # Test simplification
    bu = BaseUnits({'g':(3,2), 'k:m': 3, '[m_p]': (3,1)})
    assert str(bu) == "BaseUnits(g=3:2 km=3 [m_p]=3)"
    
def test_base():
    
    base = BaseUnits({'k:m':3,'g':2}).base()
    assert base.magnitude       == 1.0e9
    assert str(base.dimensions) == "Dimensions(m=3 g=2)"
    base = BaseUnits({'J':1,'s':-1}).base()
    assert base.magnitude       == 1000.0
    assert str(base.dimensions) == "Dimensions(m=2 g=1 s=-3)"
    base = BaseUnits({'W':(3,2),'s':(9,2)}).base()
    assert base.magnitude       == 31622.776601683792
    assert str(base.dimensions) == "Dimensions(m=3 g=3:2)"
    


import numpy as np
from dataclasses import dataclass, field, fields
from typing import Union

from .settings import *
from .FractionClass import Fraction
from .DimensionsClass import Dimensions
from .UnitSolver import UnitSolver

@dataclass
class Base:
    
    magnitude: float
    dimensions: Dimensions

class BaseUnits:

    baseunits: dict

    def __init__(self, baseunits: Union[str,list,dict,Dimensions]=None):
        if baseunits is None:
            self.baseunits = {}
        elif isinstance(baseunits, dict):
            self.baseunits = baseunits
        elif isinstance(baseunits, Dimensions):
            self.baseunits = baseunits.value(dtype=dict)
        elif isinstance(baseunits, (list, np.ndarray)):
            self.baseunits = Dimensions(*baseunits).value(dtype=dict)
        elif isinstance(baseunits, BaseUnits):
            self.baseunits = baseunits.baseunits
        elif isinstance(baseunits, str):
            self.baseunits = UnitSolver(baseunits).baseunits
        else:
            raise Exception("Cannot initialize BaseUnits with given argument:", baseunits)
        # convert units exponents to fractions
        for unit,exp in self.baseunits.items():
            if not isinstance(exp, Fraction):
                self.baseunits[unit] = Fraction(exp)
        # remove units with zero exponents
        for unit in list(self.baseunits.keys()):
            if self.baseunits[unit].num==0:
                del self.baseunits[unit]
        # remove units with nonzero dimensions if total dimension is zero
        base = self.base()
        if base.dimensions == Dimensions():
            zerodim = Dimensions().value()
            for unitid in list(self.baseunits.keys()):
                symbol = unitid.split(SYMBOL_UNITID)[-1] if SYMBOL_UNITID in unitid else unitid
                if UNIT_STANDARD[symbol].dimensions != zerodim:
                    del self.baseunits[unitid]

    def __str__(self):
        baseunits = []
        for unit,exp in self.baseunits.items():
            if exp.num not in [0, -0]:
                unit = unit.replace(SYMBOL_UNITID,"")
                baseunits.append(f"{unit}={str(exp)}")
        baseunits = " ".join(baseunits)
        return f"BaseUnits({baseunits})"

    def __repr__(self):
        baseunits = []
        for unit,exp in self.baseunits.items():
            if exp.num not in [0, -0]:
                unit = unit.replace(SYMBOL_UNITID,"")
                baseunits.append(f"{unit}={str(exp)}")
        baseunits = " ".join(baseunits)
        return f"BaseUnits({baseunits})"

    def __add__(self, other):
        baseunits = dict(self.baseunits)
        for unit,exp in other.baseunits.items():
            baseunits[unit] = baseunits[unit]+exp if unit in baseunits else exp
        return BaseUnits(baseunits)
    
    def __sub__(self, other):
        baseunits = dict(self.baseunits)
        for unit,exp in other.baseunits.items():
            baseunits[unit] = baseunits[unit]-exp if unit in baseunits else -exp
        return BaseUnits(baseunits)

    def __mul__(self, other):
        baseunits = dict(self.baseunits)
        for unit,exp in self.baseunits.items():
            baseunits[unit] *= other
        return BaseUnits(baseunits)

    def __truediv__(self, div):
        baseunits = dict(self.baseunits)
        for unit,exp in self.baseunits.items():
            baseunits[unit] /= div
        return BaseUnits(baseunits)
    
    def __eq__(self, other):
        if len(self.baseunits) != len(other.baseunits):
            return False
        for unit,exp in self.baseunits.items():
            if unit not in other.baseunits:
                return False
            elif not other.baseunits[unit] == exp:
                return False
        return True
    
    def base(self, unitid=None):
        def unit_base(unitid):
            exp = self.baseunits[unitid]
            if SYMBOL_UNITID in unitid:
                prefix, base = unitid.split(SYMBOL_UNITID)
                magnitude  = (UNIT_PREFIXES[prefix].magnitude*UNIT_STANDARD[base].magnitude) ** exp.value(dtype=float)
            else:
                prefix, base = '', unitid
                magnitude  = UNIT_STANDARD[base].magnitude ** exp.value(dtype=float)
            dimensions = Dimensions(*UNIT_STANDARD[base].dimensions)*exp
            return magnitude, dimensions
        if unitid:
            magnitude, dimensions = unit_base(unitid)
        else:
            magnitude = 1
            dimensions = Dimensions()
            for unitid in self.baseunits.keys():
                mag, dim = unit_base(unitid)
                magnitude *= mag
                dimensions += dim
        return Base(magnitude, dimensions)
    
    def expression(self):
        units = []
        for unitid,exp in self.baseunits.items():
            symbol = unitid.replace(SYMBOL_UNITID,'')
            if exp.num==1 and exp.den==1:
                units.append(symbol)
            else:
                units.append(f"{symbol}{exp}")
        if units:
            return SYMBOL_MULTIPLY.join(units)
        else:
            return None
        
    def value(self):
        baseunits = {}
        for unit,exp in self.baseunits.items():
            baseunits[unit] = exp.value()
        return baseunits



import numpy as np

class Converter:

    def __new__(cls, magnitude1, baseunits1, baseunits2):
        obj = object.__new__(cls)
        if method := obj.method(baseunits1, baseunits2):
            if not hasattr(obj,method[0]):
                raise Exception('Conversion method is not implemented:', method[0])
            obj.method = method
            return obj
        else:
            return None

    def __init__(self, magnitude1, baseunits1, baseunits2):
        base1 = baseunits1.base()
        base2 = baseunits2.base()
        self.magnitude = getattr(self,self.method[0])(magnitude1 * base1.magnitude, *self.method[1:]) / base2.magnitude

class StandardConverter(Converter):
    
    def method(self, baseunits1, baseunits2):
        base1 = baseunits1.base()
        base2 = baseunits2.base()
        if base1.dimensions==base2.dimensions:
            return (f"convert_linear",)
        elif -base1.dimensions==base2.dimensions:
            return (f"convert_inversed",)
        else:
            return False
            
    def convert_inversed(self, value):
        return 1/value

    def convert_linear(self, value):
        return value

class SingleUnitConverter(Converter):

    def method(self, baseunits1, baseunits2):
        baseunits1 = baseunits1.value()
        baseunits2 = baseunits2.value()
        convert = False
        for unitid in baseunits1.keys():
            if unitid.split(':')[-1] in self.process:
                convert = True
        for unitid in baseunits2.keys():
            if unitid.split(':')[-1] in self.process:
                convert = True
        if convert:
            if len(baseunits1)!=1 or len(baseunits2)!=1:
                raise Exception("Only simple units can be converted between each other:",
                                baseunits1, baseunits2)
            symbol1 = list(baseunits1.keys())[0].split(':')[-1]
            symbol2 = list(baseunits2.keys())[0].split(':')[-1]
            return (f"convert_{symbol1}_{symbol2}",)
        else:
            return False

class TemperatureConverter(SingleUnitConverter):
    
    process = ['Cel','degF']
    
    def convert_K_Cel(self, value):
        return value-273.15
        
    def convert_K_degF(self, value):
        return (value-273.15)*9/5+32
        
    def convert_degR_degF(self, value):
        return value*9/5-459.67
        
    def convert_degR_Cel(self, value):
        return (value*9/5-491.67)*5/9
        
    def convert_Cel_K(self, value):
        return value+273.15
        
    def convert_Cel_degF(self, value):
        return (value*9/5)+32
        
    def convert_Cel_degR(self, value):
        return ((value*9/5)+491.67)*5/9
        
    def convert_degF_K(self, value):
        return ((value-32)*5/9)+273.15
    
    def convert_degF_Cel(self, value):
        return (value-32)*5/9
        
    def convert_degF_degR(self, value):
        return (value+459.67)*5/9
        
        
class LogarithmicConverter(SingleUnitConverter):

    process: list = ['Np','B','Bm','BmW','BW','BV','BuV']
    
    def method(self, baseunits1, baseunits2):
        conversions = {
            # power ratios
            'convert_PR_B':  ("convert_Ratio_B",  1,   1),
            'convert_B_PR':  ("convert_B_Ratio",  1,   1),
            'convert_PR_Np': ("convert_Ratio_Np", 0.5, 1),
            'convert_Np_PR': ("convert_Np_Ratio", 0.5, 1),
            'convert_W_Bm':  ("convert_Ratio_B",  1,   1),
            'convert_W_BmW': ("convert_Ratio_B",  1,   1),
            'convert_Bm_W':  ("convert_B_Ratio",  1,   1),
            'convert_BmW_W': ("convert_B_Ratio",  1,   1),
            'convert_W_BW':  ("convert_Ratio_B",  1,   1e-3),
            'convert_BW_W':  ("convert_B_Ratio",  1,   1e3),
            # amplitude ratios
            'convert_AR_B':  ("convert_Ratio_B",  2,   1),
            'convert_B_AR':  ("convert_B_Ratio",  2,   1),
            'convert_AR_Np': ("convert_Ratio_Np", 1,   1),
            'convert_Np_AR': ("convert_Np_Ratio", 1,   1),
            'convert_V_BV':  ("convert_Ratio_B",  2,   1e-3),
            'convert_BV_V':  ("convert_B_Ratio",  2,   1e3),
            'convert_V_BuV': ("convert_Ratio_B",  2,   1e3),
            'convert_BuV_V': ("convert_B_Ratio",  2,   1e-3),
        }
        method = super().method(baseunits1, baseunits2)
        if method and method[0] in conversions:
            return conversions[method[0]]
        else:
            return method

    def convert_B_B(self, value):
        return value
        
    def convert_B_Np(self, value):
        return 1.151277918*value
        
    def convert_Np_B(self, value):
        return value/1.151277918
        
    def convert_Ratio_B(self, value, exp, conv):
        return exp*np.log10(value*conv)
        
    def convert_B_Ratio(self, value, exp, conv):
        return np.power(10,value/exp)*conv
        
    def convert_Ratio_Np(self, value, exp, conv):
        return exp*np.log(value*conv)
        
    def convert_Np_Ratio(self, value, exp, conv):
        return np.exp(value/exp)*conv
        


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import sys
sys.path.insert(0,"../../src")
import scinumtools

import toml
data = toml.load("../../pyproject.toml")

project = data['project']['name']
copyright = '2023, '+data['project']['authors'][0]['name']
author = data['project']['authors'][0]['name']
release = data['project']['version']
version = data['project']['version']

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["sphinx.ext.autodoc", 'sphinx_rtd_theme']

templates_path = ['_templates']
exclude_patterns = []



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']



import os
import numpy as np

def CachedFunction(file_cache):
    """ Decorator function that cashes data produced by a function

    :param str file_cache: Name of the file that will store cached data
    """
    def wrapped(func):
        def inner(*args, **kwargs):
            if not os.path.isfile(file_cache):
                np.save(file_cache, func(*args, **kwargs))
            data = np.load(file_cache, allow_pickle=True)
            return data.item() if data.ndim==0 else data
        return inner
    return wrapped



import numpy as np
import pandas as pd
from textwrap import dedent
import sys
sys.path.insert(0, 'src')

import scinumtools as snt

def test_parameter_list():
    def test(params):
        assert params[0].a == 1
        assert params[1]['a'] == 4
        for param in params:
            assert param['a'] in [1,4]
        for index,settings in params.items():
            assert param['b'] in [2,5]        
    # append values
    with snt.ParameterTable(['a','b','c']) as params:
        params.append([1, 2, 3])
        params.append([4, 5, 6])
        test(params)
    # direct value insertion
    params = snt.ParameterTable(['a','b','c'],[
        [1, 2, 3],
        [4, 5, 6]
    ])
    test(params)

def test_parameter_dict():
    def test(params):
        assert params['d'].a == 1
        assert params['e']['a'] == 4
        for param in params:
            assert param['a'] in [1,4]
        for index,settings in params.items():
            assert param['b'] in [2,5]
    # append values
    with snt.ParameterTable(['a','b','c'], keys=True) as params:
        params['d'] = [1, 2, 3]
        params.append( 'e', [4, 5, 6] )
        test(params)
    # direct value insertion
    params = snt.ParameterTable(['a','b','c'],{
        'd': [1, 2, 3],
        'e': [4, 5, 6]
    }, keys=True)
    test(params)



from PIL import Image
import numpy as np
import matplotlib.pyplot as plt

class ThumbnailImage:
    """ Create a thumbnail from an image

    :param tuple extent: Extent of an image (xmin, xmax, ymin, ymax)
    :param data: Image data
    """
    
    extent: list
    im: Image = None
    
    def __init__(self, extent, data, mode='F'):
        self.im = Image.fromarray(data).convert(mode)
        self.extent = extent
        self.xratio = np.abs((extent[1]-extent[0])/self.im.size[0])
        self.yratio = np.abs((extent[3]-extent[2])/self.im.size[1])
                
    def resize(self, extent=None, res=None, bgcolor=0):
        """ Resize image

        :param tuple extent: New extent of an image (xmin, xmax, ymin, ymax)
        :param tuple res: Resolution of a new image
        :param bgcolor: Color of the thumbnail padding (loat or a tuple)
        """
        if extent:
            xpix = int(np.round((extent[1]-extent[0])/self.xratio))
            ypix = int(np.round((extent[3]-extent[2])/self.yratio))
            layer = ThumbnailImage(
                extent = extent,
                data = np.full((ypix,xpix),bgcolor).astype(np.asarray(self.im).dtype),
                mode = self.im.mode,
            )
            pos = [int((layer.im.size[i]-self.im.size[i])/2) for i in range(2)]
            layer.im.paste(self.im, pos)
        if res and extent:
            layer.im = layer.im.resize(res)
            return layer
        else:
            self.im = self.im.resize(res)
            return self
        
    def draw(self, ax=None, **kwargs):
        """ Draw current image on an axis

        :param ax: Matplotlib axis
        """
        if not ax: ax=plt
        return ax.imshow(np.asarray(self.im), extent=self.extent, origin='lower', **kwargs)



import numpy as np
import pytest
from math import isclose
import os
import sys
sys.path.insert(0, 'src')

from scinumtools.units import *
    
def test_dimensions():

    # Test simplification
    dims = Dimensions(m=(5,-2), g=3, s=1, cd=(-0,3), K=(34,1), rad=(18,12))
    assert str(dims) == "Dimensions(m=-5:2 g=3 s=1 K=34 rad=3:2)"

def test_arithmetics():
    
    dims1 = Dimensions(m=3, g=(3,2))
    dims2 = Dimensions(m=2, g=(4,7))
    assert not dims1==dims2
    assert str(dims1+dims2) == "Dimensions(m=5 g=29:14)"
    assert str(dims1-dims2) == "Dimensions(m=1 g=13:14)"
    assert str(dims1*2)     == "Dimensions(m=6 g=3)"
    assert str(dims2*0.5)   == "Dimensions(m=1 g=2:7)"

def test_values():
    
    value = [3, (3,2), 0, 0, 0, 0, 0, 0]
    dims = Dimensions(*value)
    assert str(dims) == "Dimensions(m=3 g=3:2)" 
    assert dims.value() == value
    assert dims.value(dtype=dict) == {'m': 3, 'g':(3,2)}



import numpy as np
import pytest
from math import isclose
import os
import sys
sys.path.insert(0, 'src')

from scinumtools.units import *
from scinumtools.units.settings import *
    
def test_unit_list():
    
    assert str(Unit())
    assert str(Constant())

    #print(Constant())
    #exit(1)
    
def test_unique_names():
    
    units = list(UNIT_STANDARD.keys())
    for symbol, unit in UNIT_STANDARD.items():
        if isinstance(unit.prefixes, list):
            for prefix in unit.prefixes:
                assert prefix in UNIT_PREFIXES
                units.append(f"{prefix}{symbol}")
        elif unit.prefixes is True:
            for prefix in UNIT_PREFIXES.keys():
                units.append(f"{prefix}{symbol}")
    #print(units)
    #exit(1)
    units.sort()
    seen = set()
    dupes = [x for x in units if x in seen or seen.add(x)]    
    assert len(dupes)==0
    
def test_prefixes():
    
    # unit can have any prefix
    assert str(Quantity(1, 'km'))    == "Quantity(1.000e+00 km)"
    
    # unit cannot have prefixes
    with pytest.raises(Exception) as excinfo:
        Quantity(1, 'kCel')
    assert excinfo.value.args[0]=="Unit cannot have any prefixes:"
    assert excinfo.value.args[1]=="Cel"
    
    # unit can have only selected prefixes
    with pytest.raises(Exception) as excinfo:
        Quantity(1, 'mpc')
    assert excinfo.value.args[0]=="Unit can have only following prefixes:"
    assert excinfo.value.args[1]==['k', 'M', 'G', 'T']
    

def test_definitions():
    
    for symbol, unit in UNIT_STANDARD.items():
        if not isinstance(unit.definition,str):
            continue
        q = Quantity(1, unit.definition)
        base = q.baseunits.base()
        magnitude = q.magnitude*base.magnitude
        if base.dimensions.value(dtype=list) != unit.dimensions or not isclose(magnitude,  unit.magnitude, rel_tol=MAGNITUDE_PRECISION):
            print(q, symbol, unit.definition, unit.magnitude)
        assert isclose(magnitude,  unit.magnitude, rel_tol=MAGNITUDE_PRECISION)
        assert base.dimensions.value(dtype=list) == unit.dimensions
        
    for symbol, unit in UNIT_PREFIXES.items():
        if not isinstance(unit.definition,str):
            continue
        q = Quantity(1, unit.definition)
        base = q.baseunits.base()
        assert isclose(q.magnitude,  unit.magnitude, rel_tol=MAGNITUDE_PRECISION)
        assert base.dimensions.value(dtype=list) == unit.dimensions
        


import time
import numpy as np

from .RowCollectorClass import RowCollector

class Stopwatch:
    """ Helper class that measures time of various processes in the code

    All nodes have to be closed in an oposite order as started.
    """

    _watches: dict
    _nodes: list

    def __enter__(self):
        return self
    
    def __exit__(self, type, value, tb):
        pass
    
    def __init__(self):
        self._nodes = []
        self._watches = {
            'stopwatch': {'laps':0, 'time': 0}
        }
    
    def start(self, name:str):
        """ Start a new measuring node

        :param str name: Name of a node
        """
        start = time.time()
        self._nodes.append(name)
        node = '/'.join(self._nodes)
        if node not in self._watches:
            self._watches[node] = {
                'start': 0.,
                'laps': 0,
                'time': 0.
            }
        self._watches[node]['start'] = time.time()
        self._watches[node]['laps'] += 1
        self._watches['stopwatch']['time'] += time.time()-start
        self._watches['stopwatch']['laps'] += 1

    def stop(self, name:str):
        """ Stop existing node

        :param str name: Name of a node
        """
        start = time.time()
        if self._nodes[-1]!=name:
            raise Exception(f"Stopwatch node '{self._nodes[-1]}' was not closed!")
        node = '/'.join(self._nodes)
        self._watches[node]['time'] += time.time()-self._watches[node]['start']
        self._watches[node]['start'] = 0.
        self._watches['stopwatch']['time'] += time.time()-start
        self._watches['stopwatch']['laps'] += 1
        self._nodes.pop()

    def observer(self, name:str):
        """ Start an observer

        :param str name: Name of a node
        """
        return StopwatchObserver(name,self)
        
    def report(self):
        """ Print time report for all nodes
        """        
        lc = RowCollector(['Laps','Time','Time/Lap','Node'])
        for name,watch in self._watches.items():
            lc.append([
                watch['laps'],
                watch['time'],
                watch['time']/watch['laps'],
                name
            ])
        lc.sort('Time')
        return lc

class StopwatchObserver:

    name: str
    _stopwatch: Stopwatch
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, tb):
        self._stopwatch.stop(self.name)
        pass

    def __init__(self, name, stopwatch):
        self.name = name
        self._stopwatch = stopwatch
        self._stopwatch.start(self.name)



import numpy as np
import os
import sys
sys.path.insert(0, 'src')

import scinumtools as snt

def test_data_plot_grid_list():
    # test lists
    grid = [
        (0, 0, 0, 0),
        (1, 0, 1, 1),
        (2, 1, 0, 2),
        (3, 1, 1, 3),
        (4, 2, 0, 4),
    ]
    dpg = snt.DataPlotGrid(list(range(5)), 2)
    for i, m, n, v in dpg.items():
        assert grid[i] == (i, m, n, v)
    for i, m, n in dpg.items(missing=True):
        assert (i, m, n) == (5, 2, 1)
    grid = [
        (0, 0, 0, 0),
        (1, 1, 0, 1),
        (2, 2, 0, 2),
        (3, 0, 1, 3),
        (4, 1, 1, 4),
    ]
    for i, m, n, v in dpg.items(transpose=True):
        assert grid[i] == (i, m, n, v)
    for i, m, n in dpg.items(transpose=True, missing=True):
        assert (i, m, n) == (5, 2, 1)

def test_data_plot_grid_dict():        
    # test dictionaries
    dpg = snt.DataPlotGrid(dict(
        a = 0,
        b = 1,
        c = 2,
        d = 3,
        e = 4
    ), 2)
    grid = [
        (0, 0, 0, 'a', 0),
        (1, 0, 1, 'b', 1),
        (2, 1, 0, 'c', 2),
        (3, 1, 1, 'd', 3),
        (4, 2, 0, 'e', 4),
    ]
    for i,m,n,k,v in dpg.items():
        assert grid[i] == (i,m,n,k,v)
    for i,m,n in dpg.items(missing=True):
        assert (5, 2, 1) == (i,m,n)
    grid = [
        (0, 0, 0, 'a', 0),
        (1, 1, 0, 'b', 1),
        (2, 2, 0, 'c', 2),
        (3, 0, 1, 'd', 3),
        (4, 1, 1, 'e', 4),
    ]
    for i,m,n,k,v in dpg.items(transpose=True):
        assert grid[i] == (i,m,n,k,v)
    for i,m,n in dpg.items(transpose=True,missing=True):
        assert (5, 2, 1) == (i,m,n)
        


import numpy as np

from .settings import *
from .QuantityClass import Quantity

class Unit:
    
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass
        
    def __new__(cls, unit=None):
        if unit:
            return Quantity(1,str(unit))
        else:
            return object.__new__(cls)
    
    def __getattr__(self, unit):
        return Quantity(1,str(unit))
        
    def __str__(self):
        return Unit._list()

    def __rep__(self):
        return Unit._list()

    @staticmethod
    def _list():
        def get_pref(prefixes):
            if isinstance(prefixes,list):
                return ", ".join(prefixes)
            elif prefixes is True:
                return 'all'
            else:
                return ''
        
        sw, uw, dw, pw = 8, 20, 20, 15
        text =  "Units\n"
        text += "\nPrefixes:\n\n"
        text += f"{'Symbol':{sw}s} | {'Prefix':{uw}s} | {'Definition':{dw}s}\n"
        text += "-"*sw+"-+-"+"-"*uw+"-+-"+"-"*dw+"\n"
        for symbol, prefix in UNIT_PREFIXES.items():
            text += f"{symbol:{sw}s} | {prefix['name']:{uw}s} | {prefix['definition']:{dw}s}\n"
        text += "\nBase units:\n\n"
        text += f"{'Symbol':{sw}s} | {'Unit':{uw}s} | {'Prefixes':{pw}s}\n"
        text += "-"*sw+"-+-"+"-"*uw+"-+-"+"-"*pw+"\n"
        for symbol, unit in UNIT_STANDARD.items():
            if not symbol.startswith('[') and unit['definition'] is None:
                pref = get_pref(unit['prefixes'])
                text += f"{symbol:{sw}s} | {unit['name']:{uw}s} | {pref:{pw}s}\n"         
        text += "\nDerived units:\n\n"
        text += f"{'Symbol':{sw}s} | {'Unit':{uw}s} | {'Prefixes':{pw}s} | {'Definition':{dw}s}\n"
        text += "-"*sw+"-+-"+"-"*uw+"-+-"+"-"*pw+"-+-"+"-"*dw+"\n"
        for symbol, unit in UNIT_STANDARD.items():
            if symbol.startswith('['): continue
            if not isinstance(unit['definition'], str): continue
            if symbol=='degR': continue
            pref = get_pref(unit['prefixes'])
            text += f"{symbol:{sw}s} | {unit['name']:{uw}s} | {pref:{pw}s} | {unit['definition']:{dw}s}\n"
        text += "\nTemperature units:\n\n"
        text += f"{'Symbol':{sw}s} | {'Unit':{uw}s} | {'Definition':{dw}s}\n"
        text += "-"*sw+"-+-"+"-"*uw+"-+-"+"-"*dw+"\n"
        for symbol, unit in UNIT_STANDARD.items():
            if symbol not in ['Cel', 'degR', 'degF']:
                continue
            defn = unit['definition'] if isinstance(unit['definition'],str) else 'fn(K)'
            text += f"{symbol:{sw}s} | {unit['name']:{uw}s} | {defn:{dw}s}\n"
        text += "\nLogarithmic units:\n\n"
        text += f"{'Symbol':{sw}s} | {'Unit':{uw}s} | {'Definition':{dw}s}\n"
        text += "-"*sw+"-+-"+"-"*uw+"-+-"+"-"*dw+"\n"
        for symbol, unit in UNIT_STANDARD.items():
            if symbol not in ['Np','B','Bm','BmW','BW','BV','BuV']:
                continue
            defn = unit['definition'] if isinstance(unit['definition'],str) else 'fn(K)'
            text += f"{symbol:{sw}s} | {unit['name']:{uw}s} | {defn:{dw}s}\n"
        return text
    
    @staticmethod
    def list():
        print(Unit._list())



from scinumtools.CachedFunctionClass import CachedFunction
from scinumtools.NormalizeDataClass import NormalizeData
from scinumtools.DataPlotGridClass import DataPlotGrid
from scinumtools.ThumbnailImageClass import ThumbnailImage
from scinumtools.DataCombinationClass import DataCombination
from scinumtools.StopwatchClass import Stopwatch
from scinumtools.RowCollectorClass import RowCollector
from scinumtools.ParameterTableClass import ParameterTable
from scinumtools.ProgressBarClass import ProgressBar

import scinumtools.units as units
import scinumtools.solver as solver